#!/usr/bin/env python
#
#   export-index.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Index server maintenence script

Writes the contents of the index (in an XML format) to stdout.

usage: export-index.py database
"""

import os
import sys
from xml.etree import cElementTree as etree

import progtools

from zfilenet.indexserver import index
from zfilenet.util import data2hex, format_datetime

__author__ = 'Ross Light'
__date__ = 'June 15, 2008'
__all__ = ['ZFN_INDEX_NAMESPACE',
           'zfnxml',
           'serialize_indexobj',
           'main',]

ZFN_INDEX_NAMESPACE = "http://launchpad.net/zfilenet/ns/index"

def zfnxml(tag):
    """
    Converts a tag to the ZFN index namespace.
    
    :Parameters:
        tag : str
            The name of the element
    :Returns: The qualified tag name
    :ReturnType: str
    """
    return '{%s}%s' % (ZFN_INDEX_NAMESPACE, tag)

def serialize_indexobj(obj):
    """
    Serializes an object as an XML element tree.
    
    :Parameters:
        obj : `zfilenet.indexserver.index.IndexObject`
            The object to serialize
    :Returns: The object's XML representation
    :ReturnType: xml.etree.ElementTree.Element
    """
    # Determine correct tag
    if isinstance(obj, index.File):
        tag = zfnxml('file')
    elif isinstance(obj, index.Directory):
        tag = zfnxml('directory')
    elif isinstance(obj, index.NullObject):
        # Null objects should not be serialized
        return
    else:
        raise TypeError("Unrecognized index object")
    # Serialize common data
    elem = etree.Element(tag, uuid=str(obj.uuid))
    name_elem = etree.SubElement(elem, zfnxml("name"))
    name_elem.text = obj.name
    # Serialize dates
    date_elem = etree.SubElement(elem, zfnxml("created"))
    date_elem.text = format_datetime(obj.created)
    date_elem = etree.SubElement(elem, zfnxml("last-modified"))
    date_elem.text = format_datetime(obj.modified)
    date_elem = etree.SubElement(elem, zfnxml("last-accessed"))
    date_elem.text = format_datetime(obj.accessed)
    # Serialize specific data
    if isinstance(obj, index.File):
        length_elem = etree.SubElement(elem, zfnxml("length"))
        length_elem.text = str(obj.length)
        checksum_elem = etree.SubElement(elem, zfnxml("checksum"),
                                         method="SHA1")
        checksum_elem.text = data2hex(obj.sha1_hash)
        blob_list_elem = etree.SubElement(elem, zfnxml("blob-list"))
        for blob_id in obj.blob_list:
            blob_elem = etree.SubElement(blob_list_elem, zfnxml("blob"))
            blob_elem.text = str(blob_id)
    elif isinstance(obj, index.Directory):
        for child in obj.get_children():
            elem.append(serialize_indexobj(child))
    # Return result
    return elem

@progtools.reportexits
def main(args=None):
    # Get arguments
    if args is None:
        args = sys.argv[1:]
    if len(args) != 1:
        raise progtools.UsageError()
    # Connect to index
    if not os.path.exists(args[0]):
        print >> sys.stderr, "%s: Could not find database file" % \
            (progtools.get_program_name())
        return 1
    index_db = index.connect(args[0])
    session = index.create_session(bind=index_db)
    # Create document
    root = etree.Element(zfnxml("zfn-index"), version='0.2')
    doc = etree.ElementTree(root)
    root_uuid = str(index.get_root(session).uuid)
    root_uuid_elem = etree.SubElement(root, zfnxml("root-uuid"))
    root_uuid_elem.text = root_uuid
    # Find top-level objects
    top_levels = session.query(index.IndexObject).filter_by(_parent=None)
    for obj in top_levels:
        root.append(serialize_indexobj(obj))
    # Write to output file
    output_file = progtools.open_output_file('-', 'wb')
    doc.write(output_file)
    output_file.close()
    # Exit successfully
    session.close()
    return 0

if __name__ == '__main__':
    sys.exit(main())
