#!/usr/bin/env python
#
#   dispatch.py
#   pinet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""The dispatch system."""

__author__ = 'Ross Light'
__date__ = 'September 6, 2007'
__docformat__ = 'reStructuredText'
__all__ = ['MethodNotFound',
           'expose',
           'get_exposed_methods',
           'Dispatcher',]

class MethodNotFound(Exception):
    """Exception raised when a dispatched method is not found."""
    
    def __init__(self, method):
        super(MethodNotFound, self).__init__()
        self.method = method
    
    def __repr__(self):
        return "MethodNotFound(%r)" % (self.method)
    
    def __str__(self):
        return "Method '%s' not found" % self.method

def expose(name=None, include_client=False):
    """
    Exposes a function for network use.
    
    This function is intended to be used as a decorator, and the methods it is
    invoked on are intended to be retrieved with `getExposedMethods`.
    
    :Parameters:
        name : str
            The exposed name of the function
    :Keywords:
        include_client : bool
            Whether the first argument of the method should be the client
    :Returns: A wrapper function to call on the method
    :ReturnType: function
    """
    def wrap_func(func):
        func._pinet_dispatch_expose = True
        if name:
            func._pinet_dispatch_name = str(name)
        else:
            func._pinet_dispatch_name = func.__name__
        func._pinet_dispatch_include_client = bool(include_client)
        return func
    return wrap_func

def get_exposed_methods(obj):
    """
    Finds all exposed methods of an object.
    
    :Parameters:
        obj
            Object to search for methods.
    :Returns: The exposed methods
    :ReturnType: dict
    """
    function_dict = {}
    for attr_name in dir(obj):
        value = getattr(obj, attr_name)
        if getattr(value, '_pinet_dispatch_expose', False):
            name = getattr(value, '_pinet_dispatch_name', attr_name)
            function_dict[name] = value
    return function_dict

class Dispatcher(object):
    """An object that dispatches functions based on exposed names."""
    
    def __init__(self):
        self.functions = {}
    
    def register_function(self, func, name=None, include_client=False):
        """
        Registers a single function.
        
        :Parameters:
            name : str
                The name to expose the function as.  If no name is given, the
                function's ``__name__`` attribute is used.
        :Keywords:
            include_client : bool
                Whether the first argument of the function should be the client
        """
        if name is None:
            name = func.__name__
        self.functions[name] = (func, bool(include_client))
    
    def register_object(self, obj, prefix=None):
        """
        Register all of the exposed methods of an object.
        
        :Parameters:
            obj
                The object whose exposed methods should be registered.
            prefix : str
                The pseudo-namespace to put the methods into.
        """
        methods = get_exposed_methods(obj)
        for name, func in methods.iteritems():
            include_client = func._pinet_dispatch_include_client
            if prefix is not None:
                name = prefix + '.' + name
            self.register_function(func, name,
                                   include_client=include_client)
    
    def dispatch(self, client, method, params):
        """
        Dispatch a function.
        
        :Parameters:
            client : `JSON_RPCProtocol`
                The protocol that requested the dispatch.
            method : str
                The name of the function to call.
            params : list or dict
                Arguments to pass to the function.
        :Returns: The result of the dispatched function.
        """
        try:
            # Find function
            func, include_client = self.functions[method]
        except KeyError:
            # Method not found
            raise MethodNotFound(method)
        else:
            # Dispatch function
            if include_client:
                if isinstance(params, dict):
                    return func(client, **params)
                else:
                    return func(client, *params)
            else:
                if isinstance(params, dict):
                    return func(**params)
                else:
                    return func(*params)
