#!/usr/bin/env python
#
#   jsonrpc.py
#   pinet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
A light-weight asynchronous network system.

This module makes extensive use of Twisted_, an asynchronous networking
package.  The communication protocol is JSON-RPC, version 2.0 of the
specification.  As such, the system is probably not optimized for high-volume
messaging or extremely low-bandwidth situations.  However, for most purposes,
JSON-RPC is compact enough to be a viable option.

Features supported by this module:

* Peer-based connections
* Asynchronous request and response
* Notifications
* Python proxy objects (in the spirit of xmlrpclib)

.. _Twisted: http://twistedmatrix.com/
"""

import socket
from uuid import UUID, uuid4

import simplejson
from twisted.internet import defer
from twisted.internet.protocol import (Protocol,
                                       Factory,
                                       ServerFactory,
                                       ClientFactory)

from pinet import dispatch

__author__ = 'Ross Light'
__date__ = 'August 8, 2007'
__docformat__ = 'reStructuredText'
__all__ = ['NoPeerConnection',
           'RPCError',
           'generate_id',
           'JSON_RPCProtocol',
           'JSON_RPCFactory',
           'ServerProxy',]

class NoPeerConnection(ValueError):
    """Raised when the requested connection is gone."""

class RPCError(Exception):
    """
    Raised when there is an error during RPC.
    
    :IVariables:
        code : int
            The error code
        message : str
            A brief description of the error
        data : dict
            Any additional data provided with the error
    """
    ERRCODE_PARSE_ERROR = -32700
    ERRCODE_INVALID_REQUEST = -32600
    ERRCODE_METHOD_NOT_FOUND = -32601
    ERRCODE_INVALID_PARAMS = -32602
    ERRCODE_INTERNAL_ERROR = -32603
    
    @classmethod
    def fromdict(cls, value):
        """
        Interprets an RPC error from a dictionary.
        
        :Parameters:
            value : dict
                The dictionary to interpret
        :Returns: The equivalent error
        :ReturnType: RPCError
        """
        # Get required values
        try:
            code = value['code']
        except KeyError:
            raise ValueError("Error object has no code member")
        try:
            message = value['message']
        except KeyError:
            raise ValueError("Error object has no message")
        # Data may be omitted
        data = value.get('data')
        # Create error
        return cls(code, message, data)
    
    def __init__(self, code, message, data=None):
        super(RPCError, self).__init__()
        self.code = code
        self.message = message
        self.data = data
    
    def __repr__(self):
        return "RPCError(%r, %r, %r)" % (self.code, self.message, self.data)
    
    def __str__(self):
        return str(self.message)
    
    def todict(self):
        """
        Converts an RPC error into a dictionary.
        
        :Returns: The equivalent dictionary
        :ReturnType: dict
        """
        return {'code': self.code,
                'message': self.message,
                'data': self.data,}

def generate_id():
    """
    Create a unique identifier for a JSON-RPC request.
    
    :Parameters:
        length : int
            The length of the identifier
    :Returns: A unique identifier
    :ReturnType: str
    """
    return uuid4().hex

class JSON_RPCProtocol(Protocol):
    """A JSON-RPC connection."""
    
    def connectionMade(self):
        self.data_buffer = ""
        self.pending_responses = {}
        self.factory.add_peer(self)
    
    def connectionLost(self, reason):
        self.factory.remove_peer(self)
    
    def dataReceived(self, data):
        decoder = self.factory.decoder
        data = self.data_buffer + data
        while data:
            # Find the start of a dictionary
            try:
                position = data.find("{")
            except ValueError:
                break
            data = data[position:]
            # Try to decode JSON
            try:
                (message, position) = decoder.raw_decode(data)
            except ValueError:
                break
            else:
                data = data[position:]
                self.process_object(message)
        self.data_buffer = data
    
    def process_object(self, message):
        """
        Handle a complete object sent over the connection.
        
        The default implementation dispatches the object to `processRequest`,
        `processNotification`, or `processResponse` depending on the kind of
        object it is.
        
        :Parameters:
            message : dict
                The object received over the network
        """
        # Verify we're using JSON-RPC 2.0
        if message.get('jsonrpc') != "2.0":
            # Bad object
            error = RPCError(RPCError.ERRCODE_INVALID_REQUEST,
                             "Not JSON-RPC 2.0.")
            self.send_response(None, error)
            return
        # Process as normal
        if 'method' in message and 'id' in message:
            # Request
            self.process_request(message['id'],
                                 message['method'],
                                 message.get('params', []))
        elif 'method' in message and 'id' not in message:
            # Notification
            self.process_notification(message['method'],
                                      message.get('params', []))
        elif 'result' in message or 'error' in message:
            # Response
            if 'result' in message and 'error' in message:
                # Both a result and an error; INVALID JSON-RPC!
                error = RPCError(RPCError.ERRCODE_INVALID_REQUEST,
                                 "Invalid Response.")
                self.send_response(None, error)
                return
            else:
                # Valid response
                self.process_response(message['id'],
                                      message.get('result'),
                                      message.get('error'),)
        else:
            # Bad request
            error = RPCError(RPCError.ERRCODE_INVALID_REQUEST,
                             "Invalid Request.")
            self.send_response(None, error)
    
    def process_request(self, request_id, method, params):
        """
        Handle a function request.
        
        :Parameters:
            request_id
                The request's identifier
            method : str
                The name of the method to call
            params : list
                The list of arguments to pass to the method
        """
        try:
            return_value = self.factory.requests.dispatch(self, method, params)
        except RPCError, e:
            return_value = e
        except dispatch.MethodNotFound:
            return_value = RPCError(RPCError.ERRCODE_METHOD_NOT_FOUND,
                                    "Method not found.")
        except TypeError, e:
            return_value = RPCError(RPCError.ERRCODE_INVALID_PARAMS,
                                    "Invalid params.")
        except Exception, e:
            return_value = RPCError(RPCError.ERRCODE_INTERNAL_ERROR,
                                    "Internal JSON-RPC error.")
        self.send_response(request_id, return_value)
    
    def process_response(self, response_id, result, error):
        """
        Handle a function response.
        
        :Parameters:
            response_id
                The response's identifier
            result
                The resulting object
            error : dict
                An error dictionary, or ``None`` if the function was successful.
        """
        try:
            response_deferred = self.pending_responses.pop(response_id)
        except KeyError:
            raise Exception("Bad response received")
        else:
            if error is None:
                response_deferred.callback(result)
            else:
                response_deferred.errback(RPCError.fromdict(error))
    
    def process_notification(self, name, params):
        """
        Handle a notification.
        
        :Parameters:
            name : str
                The name of the notification
            params : list or dict
                The notification's arguments
        """
        self.factory.notifications.dispatch(self, name, params)
    
    def send_object(self, obj):
        """
        Send a raw object over the connection.
        
        :Parameters:
            obj
                The object to send.
        """
        obj = dict(obj)
        obj['jsonrpc'] = "2.0"
        encoder = self.factory.encoder
        for chunk in encoder.iterencode(obj):
            self.transport.write(chunk)
    
    def send_request(self, name, params=[]):
        """
        Send a request over the connection.
        
        This isn't intended for general use; instead, use `call`.  This is
        intended for overriding, in case you want to do something special to
        send a request.
        
        :Parameters:
            name : str
                The name of the function to call.
            params : list or dict
                Parameters to give to the function.
        :Returns: The request's identifier
        :ReturnType: str
        """
        request_id = generate_id()
        body = {'method': name,
                'params': params,
                'id': request_id,}
        self.send_object(body)
        return request_id
    
    def send_notification(self, name, params=[]):
        """
        Send a notification over the connection.
        
        This isn't intended for general use; instead, use `notify`.  This is
        intended for overriding, in case you want to do something special to
        send a notification.
        
        :Parameters:
            name : str
                The name of the function to call.
            params : list or dict
                Parameters to give to the function.
        """
        body = {'method': name,
                'params': params,}
        self.send_object(body)
    
    def send_response(self, response_id, result):
        """
        Send a notification over the connection.
        
        :Parameters:
            response_id
                The response's identifier.
            result
                The result to return.  If this is a subclass of `RPCError`,
                then it is assumed that an error occurred and the object is
                used as the error.
        """
        if isinstance(result, RPCError):
            body = {'error': result.todict(),
                    'id': response_id,}
        else:
            body = {'result': result,
                    'id': response_id,}
        self.send_object(body)
    
    def call(self, name, *args, **kw):
        """
        Call a function on the other peer.
        
        :Parameters:
            name : str
                Name of the remote method
        :ReturnType: Deferred
        """
        request_id = self.send_request(name, self._get_params(args, kw))
        new_deferred = defer.Deferred()
        self.pending_responses[request_id] = new_deferred
        return new_deferred
    
    def notify(self, name, *args, **kw):
        """
        Call a function on the other peer, without expecting a response.
        
        :Parameters:
            name : str
                Name of the remote method
        """
        if len(args) > 0 and len(kw) > 0:
            raise TypeError("Positional and keyword arguments mixed")
        self.send_notification(name, self._get_params(args, kw))
    
    def get_proxy(self):
        """
        Creates a proxy of the other peer.
        
        :ReturnType: `ServerProxy`
        """
        return ServerProxy(self)
    
    @staticmethod
    def _get_params(args, kw):
        if len(args) > 0 and len(kw) > 0:
            raise TypeError("Positional and keyword arguments mixed")
        if kw:
            return kw
        else:
            return args

class JSON_RPCFactory(ServerFactory, ClientFactory):
    """A JSON-RPC peer"""
    protocol = JSON_RPCProtocol
    
    def __init__(self,
                 requests=None,
                 notifications=None,
                 encoder=None,
                 decoder=None):
        """
        Initialize a JSON-RPC factory.
        
        :Parameters:
            requests : `pinet.dispatch.Dispatcher`
                The request dispatcher.  If not given, one is created.
            notifications : `pinet.dispatch.Dispatcher`
                The notifications dispatcher.  If not given, requests is used.
        :Keywords:
            encoder : ``simplejson.JSONEncoder``
                The encoder to use for requests
            decoder : ``simplejson.JSONDecoder``
                The decoder to use for responses
        """
        # Set up requests
        if requests is None:
            requests = dispatch.Dispatcher()
        self.requests = requests
        # Set up notifications
        if notifications is None:
            notifications = requests
        self.notifications = notifications
        # Set up peers
        self.peers = []
        # Set up JSON coding
        if encoder is None:
            encoder = simplejson.JSONEncoder(separators=(',', ':'))
        self.encoder = encoder
        if decoder is None:
            decoder = simplejson.JSONDecoder()
        self.decoder = decoder
    
    def get_peer(self, peer):
        """
        Obtain a connected peer.
        
        :Parameters:
            peer : int
                The peer index to retrieve.
        :Raises NoPeerConnection: If no such peer exists
        :Returns: The connected peer
        :ReturnType: `JSON_RPCProtocol`
        """
        if peer in self.peers:
            return peer
        else:
            try:
                return self.peers[peer]
            except (KeyError, IndexError):
                raise NoPeerConnection()
    
    def add_peer(self, peer):
        """
        Add a peer to the factory.
        
        :Parameters:
            peer : `JSON_RPCProtocol`
                The peer to add to the factory.
        """
        self.peers.append(peer)
    
    def remove_peer(self, peer):
        """
        Add a peer to the factory.
        
        :Parameters:
            peer : `JSON_RPCProtocol`
                The peer to remove from the factory.
        :Raises NoPeerConnection: If no such peer is in the factory.
        """
        try:
            self.peers.remove(peer)
        except ValueError:
            raise NoPeerConnection()
    
    def buildProtocol(self, addr):
        protocol = Factory.buildProtocol(self, addr)
        self.peers.append(protocol)
        return protocol
    
    def call(self, name, *args, **kw):
        """
        Call a function on the first peer.
        
        :See: `JSON_RPCProtocol.call`
        :ReturnType: Deferred
        """
        try:
            peer = self.get_peer(0)
        except NoPeerConnection, e:
            return defer.fail(e)
        else:
            return peer.call(name, *args, **kw)
    
    def notify(self, name, *args, **kw):
        """
        Sends a notification to all peers.
        
        :See: `JSON_RPCProtocol.notify`
        """
        for peer in self.peers:
            peer.notify(name, *args, **kw)
    
    def get_proxy(self):
        """
        Creates a proxy of the first peer.
        
        :ReturnType: `ServerProxy`
        """
        return ServerProxy((self, 0))

class ServerProxy(object):
    """
    A proxy of a JSON-RPC peer.
    
    You can call a remote function like this::
    
        proxy = factory.get_proxy()          # Get the proxy from the factory
        d = proxy.spam('eggs', 'lumberjack') # The proxy returns Deferreds
        d.addCallback(gotSpam)               # gotSpam is called with the result
        d.addErrback(spamMessedUp)           # spamMessedUp is called on error
    """
    
    __slots__ = ['_protocol', '_name']
    
    def __init__(self, protocol, name=''):
        self._protocol = protocol
        self._name = name
    
    def __getattr__(self, name):
        if self._name:
            name = self._name + '.' + name
        return type(self)(self._protocol, name)
    
    def __call__(self, *args, **kw):
        # Get real protocol
        if isinstance(self._protocol, tuple):
            factory, peer = self._protocol
            protocol = factory.get_peer(peer)
        else:
            protocol = self._protocol
        # Make remote procedure call
        if self._name:
            return protocol.call(self._name, *args, **kw)
        else:
            raise ValueError("Proxy has no name")

if __name__ == '__main__':
    import math
    from twisted.internet import reactor
    def sin(theta):
        return math.sin(math.radians(theta))
    def cos(theta):
        return math.cos(math.radians(theta))
    def tan(theta):
        return math.tan(math.radians(theta))
    factory = JSON_RPCFactory()
    factory.requests.register_function(sin)
    factory.requests.register_function(cos)
    factory.requests.register_function(tan)
    reactor.listenTCP(8000, factory)
    reactor.run()
