#!/usr/bin/env python
#
#   zfilenet/indexserver/dav.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""WebDAV access to a ZFN share."""

from __future__ import absolute_import
from base64 import b64encode, b64decode
from cgi import escape
from hashlib import sha1
import logging
import mimetypes
import posixpath
import re
from tempfile import TemporaryFile
from urllib import quote, quote_plus, unquote
from urlparse import urlparse
from uuid import UUID, uuid4
from xml.etree import ElementTree

import genshi
from genshi.template import TemplateLoader
import pkg_resources
from sqlalchemy.orm import object_session
from twisted.web import error, http, resource, server
from twisted.web.static import addSlash, FileTransfer
from twisted.web.util import redirectTo

from . import errors, index
from .. import discovery, util

__author__ = 'Ross Light'
__date__ = 'June 27, 2008'
__all__ = ['log',
           'DEFAULT_PORT',
           'DAV_NAMESPACE',
           'DAVResponse',
           'DAVProperty',
           'DirectoryListing',
           'DirectoryCreator',
           'FilePutter',
           'NullLocker',
           'StaticNamespace',
           'ZFNResource',
           'HTTPChannel',
           'ZFNSite',
           'create_server',]

log = logging.getLogger('zfilenet.indexserver.dav')
access_log = logging.getLogger('zfilenet.indexserver.dav.access')
DEFAULT_PORT = 47800
DAV_NAMESPACE = 'DAV:'

_template_dir = pkg_resources.resource_filename('zfilenet', 'templates')
_template_loader = TemplateLoader([_template_dir])

### DAV RESPONSE CODES ####

PROCESSING = 102
UNPROCESSABLE_ENTITY = 422
LOCKED = 423
FAILED_DEPENDENCY = 424

http.RESPONSES.setdefault(PROCESSING, "Processing")
http.RESPONSES.setdefault(LOCKED, "Locked")
http.RESPONSES.setdefault(UNPROCESSABLE_ENTITY, "Unprocessable Entity")
http.RESPONSES.setdefault(FAILED_DEPENDENCY, "Failed Dependency")

### FUNCTIONS ###

def _render_template(template, values=None, **kw):
    method = kw.pop('genshi_format', 'html')
    trim = kw.pop('comment_trim', False)
    if values is None:
        values = {}
    else:
        values = dict(values)
    values.update(kw)
    zfn_ns = dict(quote=quote,
                  format_file_size=util.format_file_size,
                  format_datetime=util.format_datetime,
                  static=_render_static,)
    values['zfn'] = zfn_ns
    template = _template_loader.load(template)
    result_stream = template.generate(**values)
    result_data = result_stream.render(method)
    if trim:
        start_token, end_token = '<!--BEGIN-->', '<!--END-->'
        start_index = result_data.index(start_token) + len(start_token)
        end_index = result_data.rindex(end_token)
        result_data = result_data[start_index:end_index]
    return result_data

def _render_error(code, message=None):
    return _render_template("error.html",
                            code=code,
                            meaning=http.RESPONSES[code],
                            message=message,)

def _render_static(name, mime_type=None):
    return '/zfn-static?name=%s' % (quote_plus(name))

def _send_error(request, code, message=None):
    request.setResponseCode(code)
    request.setHeader('Content-Type', 'text/html')
    return _render_error(code, message)

def _get_host(request):
    port = request.getHost().port
    return 'http://' + request.getRequestHostname() + ':' + str(port)

# TODO: Complete lock parser (we're parsing entity tags wrong)
_if_list_pattern = re.compile(
    r'(?:<([^>]*)>)?\s*'                # Tag
    r'\((Not\s*)?(?:(?:<([^>]*)>)'      # Token
    r'|(?:\[(?:W/)?\"([^\"]*)\"\]))\)') # Entity Tag
def _get_locks(request):
    header = request.getHeader('If')
    if header is None:
        return []
    value = []
    while True:
        header = header.lstrip()
        match = _if_list_pattern.match(header)
        if match is None:
            break
        resource, not_, token, etag = list(match.groups())
        not_ = bool(not_ is None)
        value.append([resource, not_, token, etag])
        header = header[match.end():]
    return [comps[2] for comps in value if comps[1] and comps[2] is not None]

def _get_mime_type(index_obj):
    mime_type, encoding = mimetypes.guess_type(index_obj.name)
    return mime_type

def _format_http_date(dt):
    return dt.strftime("%a, %d %b %Y %H:%M:%S GMT")

def _get_properties(properties, index_obj):
    # Generate all properties
    lock_support_data = """\
<D:lockentry>
     <D:locktype><D:write/></D:locktype>
     <D:lockscope><D:exclusive/></D:lockscope>
</D:lockentry>
<D:lockentry>
     <D:locktype><D:write/></D:locktype>
     <D:lockscope><D:shared/></D:lockscope>
</D:lockentry>"""
    lock_discovery_data = _render_template('lock_discovery.xml',
                                           genshi_format='xml',
                                           comment_trim=True,
                                           locks=index_obj.locks,)
    item_properties = [
        DAVProperty(DAV_NAMESPACE, 'creationdate',
                    util.format_datetime(index_obj.created)),
        DAVProperty(DAV_NAMESPACE, 'getlastmodified',
                    _format_http_date(index_obj.modified)),
        DAVProperty(DAV_NAMESPACE, 'supportedlock',
                    lock_support_data, escape_value=False),
        DAVProperty(DAV_NAMESPACE, 'lockdiscovery',
                    lock_discovery_data, escape_value=False),]
    if index_obj.is_file:
        item_properties += [
            DAVProperty(DAV_NAMESPACE, 'getcontentlength', index_obj.length),
            DAVProperty(DAV_NAMESPACE, 'getcontenttype', _get_mime_type(index_obj)),
            DAVProperty(DAV_NAMESPACE, 'resourcetype', None),
        ]
    elif index_obj.is_directory:
        item_properties += [
            DAVProperty(DAV_NAMESPACE, 'resourcetype',
                        '<D:collection/>', escape_value=False),
        ]
    # Return correct properties
    if properties in ('all', 'names'):
        return item_properties, [], []
    else:
        found, not_found, forbidden = [], [], []
        for prop in properties:
            for item_prop in item_properties:
                if item_prop.namespace == prop.namespace and \
                   item_prop.name == prop.name:
                    found.append(item_prop)
                    break
            else:
                not_found.append(prop)
        return (found, not_found, forbidden)

def _set_properties(properties, index_obj):
    successful, forbidden, conflict = [], [], []
    for prop_list in properties:
        optype, prop_list = prop_list[0], prop_list[1:]
        for prop in prop_list:
            if prop.name in ('getcontentlength', 'getcontenttype',
                             'getlastmodified', 'resourcetype',
                             'creationdate'):
                # Automatically generated properties
                conflict.append(prop)
            else:
                # We don't like other properties
                forbidden.append(prop)
    # Return results
    if forbidden or conflict:
        return [], forbidden, conflict, successful
    else:
        return successful, [], [], []

def _write_file(session, blob_server_list, path, content, locks=[]):
    BUFFER_SIZE = 8 * 1024 # 8 KiB
    BLOB_SIZE = 128 * 1024 # 128 KiB
    # Determine mode
    node = index.resolve_path(session, path)
    append = False
    if node is not None and not node.is_null:
        # File does exist, check for appending
        if node.is_directory:
            # Trying to write to a directory?  Shame.
            return False
        file_hash = sha1()
        hash_length = 0
        while content.tell() < node.length:
            data = content.read(BUFFER_SIZE)
            if not data:
                break
            update_length = min(node.length - hash_length, BUFFER_SIZE)
            file_hash.update(data[:update_length])
            hash_length += len(data)
        else:
            if file_hash.digest() == node.sha1_hash:
                append = True
        if append:
            content.seek(node.length)
            file_length = node.length
        else:
            content.seek(0)
            node.clear()
            file_hash = sha1()
            file_length = 0
    else:
        # File does not exist, create it.
        try:
            node = index.File.make(session, path, locks=locks)
        except errors.LockedError:
            raise
        except errors.Error:
            return False
        else:
            file_hash = sha1()
            file_length = 0
    # Write data
    writer = blob_server_list.get_writer()
    while True:
        data = content.read(BLOB_SIZE)
        if not data:
            break
        new_blob_id = uuid4()
        node.add_blob(new_blob_id)
        file_hash.update(data)
        file_length += len(data)
        writer.write_blob(new_blob_id, data)
    # Finish it!
    node.length = file_length
    node.sha1_hash = file_hash.digest()
    session.commit()
    return True

def _parse_lock_request(request):
    # Start result dictionary
    result = {}
    result['depth'] = request.getHeader('Depth')
    if result['depth'] is None:
        result['depth'] = 'infinity'
    if result['depth'] not in ('0', 'infinity'):
        return http.BAD_REQUEST
    # Parse request
    if request.getHeader('Content-Length') == '0':
        return http.BAD_REQUEST
    else:
        doc = ElementTree.parse(request.content)
        # TODO: Check if root is {DAV:}lockinfo
        root = doc.getroot()
        # Determine lock type
        lock_type = root.find('{DAV:}locktype')
        if lock_type is None:
            return http.BAD_REQUEST
        elif lock_type[0].tag == '{DAV:}write':
            pass
        else:
            return http.PRECONDITION_FAILED
        # Determine lock scope
        scope = root.find('{DAV:}lockscope')
        if scope is None:
            return http.BAD_REQUEST
        elif scope[0].tag == '{DAV:}exclusive':
            result['exclusive'] = True
        elif scope[0].tag == '{DAV:}shared':
            result['exclusive'] = False
        else:
            return http.BAD_REQUEST
        # Determine owner
        result['owner'] = root.find('{DAV:}owner')
    return result

class DAVResponse(object):
    def __init__(self, href, properties=None, not_found=None, forbidden=None,
                 conflict=None, failed=None):
        self.href = href
        self.properties = properties if properties is not None else []
        self.not_found = not_found if not_found is not None else []
        self.forbidden = forbidden if forbidden is not None else []
        self.conflict = conflict if conflict is not None else []
        self.failed = failed if failed is not None else []

class DAVProperty(object):
    _name_pattern = re.compile(r'^[_:a-zA-Z][-_:.a-zA-Z0-9]*$')
    
    def __init__(self, *args, **kw):
        # Get arguments
        if len(args) == 2:
            (name, value), namespace = args, None
        elif len(args) == 3:
            namespace, name, value = args
        else:
            raise TypeError("DAVProperty.__init__ takes 2-3 arguments")
        escape_value = kw.pop('escape_value', True)
        if kw:
            raise TypeError("Unrecognized keyword parameter given to "
                            "DAVProperty.__init__")
        # Initialize
        if not self._name_pattern.match(name):
            raise ValueError("Bad property name")
        self.name = name
        self.value = unicode(value) if value is not None else None
        self.namespace = namespace
        self.escape_value = bool(escape_value and self.value is not None)
    
    def __repr__(self):
        if self.namespace is None:
            return "DAVProperty(%r, %r)" % (self.name, self.value)
        else:
            return "DAVProperty(%r, %r, %r)" % \
                (self.namespace, self.name, self.value)
    
    def __str__(self):
        return unicode(self).encode()
    
    def __unicode__(self):
        return self.render()
    
    def render(self, name_only=False):
        """
        Render the property to a string.
        
        :Keywords:
            name_only : bool
                Whether to only generate the name
        :Returns: The XML data of the property
        :ReturnType: unicode
        """
        value = escape(self.value) if self.escape_value else self.value
        values = {'name': self.name,
                  'value': value,}
        if self.namespace is not None:
            values['ns'] = escape(self.namespace, quote=True)
            values['prefix'] = 'D' if self.namespace == DAV_NAMESPACE else 'N'
            format = u"<%(prefix)s:%(name)s xmlns:%(prefix)s=\"%(ns)s\""
        else:
            format = u"<%(name)s"
        if name_only or value is None:
            format += u" />"
        else:
            format += u">%(value)s</"
            if 'prefix' in values:
                format += "%(prefix)s:"
            format += "%(name)s>"
        return format % values
    
    def serialize(self, name_only=False):
        """
        Render the property to a string.
        
        :Keywords:
            name_only : bool
                Whether to only generate the name
        :Returns: The XML data of the property
        :ReturnType: unicode
        """
        stream = genshi.XML(self.render(name_only=name_only))
        # We run this through a namespace flattener to make sure we get the
        # correct namespacing.
        return stream | genshi.output.NamespaceFlattener()

class DirectoryListing(resource.Resource):
    def __init__(self, directory):
        resource.Resource.__init__(self)
        self.directory = directory
        self.isLeaf = True
    
    def render_GET(self, request):
        session = object_session(self.directory)
        request.setHeader('Content-Type', 'text/html')
        request.setHeader('Last-Modified',
                          _format_http_date(self.directory.modified))
        is_root = bool(self.directory == index.get_root(session))
        user_agent = request.getHeader('User-Agent')
        if user_agent is not None and \
           ('iPod' in user_agent or 'iPhone' in user_agent):
            template = "iphone_listing.html"
        else:
            template = "listing.html"
        return _render_template(template,
                                request=request,
                                directory=self.directory,
                                is_root=is_root,)

class DirectoryCreator(resource.Resource):
    def __init__(self, directory, name, bad_path=False):
        resource.Resource.__init__(self)
        self.isLeaf = False
        self.directory = directory
        self.name = name
        self.bad = bad_path
    
    def getChild(self, path, request):
        if path:
            return DirectoryCreator(self.directory, path, True)
        else:
            return self
    
    def render_MKCOL(self, request):
        session = object_session(self.directory)
        if self.bad:
            return _send_error(request, http.CONFLICT)
        elif self.directory.get_child(self.name) is not None:
            return _send_error(request, http.NOT_ALLOWED)
        else:
            locks = _get_locks(request)
            path = self.directory.get_path() + self.name
            log.debug("Creating directory %s from MKCOL", path)
            try:
                index.Directory.make(session, path, locks)
            except errors.LockedError:
                return _send_error(request, LOCKED)
            else:
                request.setResponseCode(http.CREATED)
                return ''

class FilePutter(resource.Resource):
    def __init__(self, blob_server_list, directory, name, bad_path=False):
        resource.Resource.__init__(self)
        self.isLeaf = False
        self.blob_server_list = blob_server_list
        self.directory = directory
        self.name = name
        self.bad = bad_path
    
    def getChild(self, path, request):
        return FilePutter(blob_server_list, self.directory, path, True)
    
    def render_PUT(self, request):
        session = object_session(self.directory)
        locks = _get_locks(request)
        if self.bad:
            return _send_error(request, http.CONFLICT)
        child = self.directory.get_child(self.name)
        if child is not None:
            if not child.is_null:
                return _send_error(request, http.NOT_ALLOWED)
            elif not child.can_access(locks):
                return _send_error(request, LOCKED, "Locked for new file")
        path = self.directory.get_path() + self.name
        log.debug("Creating file %s from PUT", path)
        try:
            success = _write_file(session, self.blob_server_list,
                                  path, request.content, locks)
        except errors.LockedError:
            return _send_error(request, LOCKED)
        else:
            if success:
                request.setResponseCode(http.CREATED)
                return ''
            else:
                return _send_error(request, http.CONFLICT)

class NullLocker(resource.Resource):
    def __init__(self, directory, name, bad_path=False):
        resource.Resource.__init__(self)
        self.isLeaf = False
        self.directory = directory
        self.name = name
        self.bad = bad_path
    
    def getChild(self, path, request):
        return NullLocker(self.directory, path, True)
    
    def render_LOCK(self, request):
        session = object_session(self.directory)
        if self.bad:
            return _send_error(request, http.CONFLICT)
        # Parse request
        request_info = _parse_lock_request(request)
        if isinstance(request_info, (int, long)):
            # Error returned, stop everything
            return _send_error(request, request_info)
        # Create null
        index_obj = index.NullObject(self.name)
        session.save(index_obj)
        # Attempt lock
        try:
            index_obj.set_name_parent(self.name, self.directory,
                                      _get_locks(request))
            new_lock = index_obj.lock(exclusive=request_info['exclusive'])
        except errors.LockedError:
            index_obj.delete(None)
            return _send_error(request, LOCKED)
        else:
            log.info("Locked null '%s' with token: %s",
                     index_obj.get_path(), new_lock.name)
            request.setResponseCode(http.OK)
            request.setHeader('Content-Type', 'text/xml')
            request.setHeader('Lock-Token', '<%s>' % new_lock.name)
            lock_data = _render_template("lock_discovery.xml",
                                         genshi_format='xml',
                                         comment_trim=True,
                                         locks=index_obj.locks)
            return """\
<?xml version=\"1.0\"?>

<D:prop xmlns:D="DAV:">
    <D:lockdiscovery>
%s
    </D:lockdiscovery>
</D:prop>
""" % lock_data

class StaticNamespace(resource.Resource):
    resources = {'arrow': 'static/arrow.png',
                 'header': 'static/header.png',
                 'logo-16': 'static/logo-16.png',
                 'logo-64': 'static/logo-64.png',
                 'parent': 'static/parent.png',
                 'common-stylesheet': 'static/common.css',
                 'zfn-stylesheet': 'static/zfn.css',
                 'zfn-stylesheet-iphone': 'static/zfn_iphone.css',}
    isLeaf = True
    
    def render_GET(self, request):
        resource_name = request.args['name'][0]
        try:
            resource_location = self.resources[resource_name]
        except KeyError:
            return _send_error(request, http.NOT_FOUND)
        else:
            fileobj = pkg_resources.resource_stream('zfilenet',
                                                    resource_location)
            # Get file size
            fileobj.seek(0, 2)
            size = fileobj.tell()
            fileobj.seek(0)
            request.setHeader('Content-Length', str(size))
            # Get file type
            mime_type, encoding = mimetypes.guess_type(resource_location)
            if mime_type:
                request.setHeader('Content-Type', mime_type)
            if encoding:
                request.setHeader('Content-Encoding', encoding)
            # Handle HEAD
            if request.method == 'HEAD':
                return ''
            # Handle GET
            FileTransfer(fileobj, size, request)
            return server.NOT_DONE_YET

class ZFNResource(resource.Resource):
    _child_not_found = error.NoResource("File not found.")
    
    def __init__(self, engine, blob_server_list, index_obj=None):
        resource.Resource.__init__(self)
        self.engine = engine
        self.blob_server_list = blob_server_list
        self.index_obj = index_obj
        self.isLeaf = bool(self.index_obj is not None and
                           not self.index_obj.is_directory)
    
    def _get_session(self):
        # Determine if we have an object session
        if self.index_obj is not None:
            session = object_session(self.index_obj)
        else:
            session = None
        # If we have no session, create one
        if session is None:
            session = index.create_session(bind=self.engine)
        return session
    
    def _get_index_obj(self, session):
        if self.index_obj is None:
            return index.get_root(session)
        else:
            return self.index_obj
    
    def _get_data(self):
        session = self._get_session()
        index_obj = self._get_index_obj(session)
        return session, index_obj
    
    def listDynamicNames(self):
        session, index_obj = self._get_data()
        if index_obj.is_directory:
            return [''] + [c.name for c in index_obj.get_children()]
        else:
            return []
    
    def getChild(self, path, request):
        session, index_obj = self._get_data()
        # Special handling for /zfn-resource
        if path == 'zfn-static' and \
           self.index_obj is None and \
           request.args:
            return StaticNamespace()
        # Now for normal children
        if index_obj.is_directory:
            if not path:
                if request.method in ('HEAD', 'GET'):
                    return DirectoryListing(index_obj)
                else:
                    return self
            else:
                child = index_obj.get_child(path)
                if child is not None:
                    return ZFNResource(self.engine, self.blob_server_list,
                                       child)
                elif request.method == 'MKCOL':
                    return DirectoryCreator(index_obj, path)
                elif request.method == 'PUT':
                    return FilePutter(self.blob_server_list, index_obj, path)
                elif request.method == 'LOCK':
                    return NullLocker(index_obj, path)
                else:
                    return self._child_not_found
        else:
            return self._child_not_found
    
    def render(self, request):
        # Log weird request
        if request.method not in ('HEAD', 'GET', 'PROPFIND'):
            log.debug("ZFN: Client attempting %s on %s",
                      request.method, request.path)
        # Test for nullness
        session, index_obj = self._get_data()
        null_operations = ('PUT', 'MKCOL', 'OPTIONS', 'PROPFIND',
                           'LOCK', 'UNLOCK')
        if index_obj.is_null and request.method not in null_operations:
            return _send_error(request, http.NOT_FOUND)
        # Do normal render
        return resource.Resource.render(self, request)
    
    def render_OPTIONS(self, request):
        request.setHeader('DAV', '2')
        return ''
    
    def render_GET(self, request):
        session, index_obj = self._get_data()
        log.debug("GETing %s", request.path)
        if index_obj.is_directory:
            # Directory without trailing slash
            return self.index_redirect(request)
        if index_obj.length == 0 or not index_obj.blob_list:
            # Empty file
            request.setHeader('Content-Type', _get_mime_type(index_obj))
            request.setHeader('Last-Modified',
                              _format_http_date(index_obj.modified))
            return ''
        # Serving regular file
        log.debug("Serving file %s (UUID %s)",
                  index_obj.name, index_obj.uuid)
        fetcher = self.blob_server_list.fetch_list(index_obj.blob_list)
        fetcher.has_blobs.addCallback(self._blobs_mapped,
                                      request, session, index_obj)
        fetcher.progress.addCallback(self._blob_fetched,
                                     request, session, index_obj)
        fetcher.finished.addCallback(self._blobs_done,
                                     request, session, index_obj)
        fetcher.start()
        return server.NOT_DONE_YET
    
    def index_redirect(self, request):
        return redirectTo(addSlash(request), request)
    
    def _blobs_mapped(self, success, request, session, index_obj):
        log.debug("Finished mapping (success: %i)", success)
        if success:
            # Write headers
            request.setHeader('Content-Type', _get_mime_type(index_obj))
            request.setHeader('Last-Modified',
                              _format_http_date(index_obj.modified))
            request.setHeader('Content-Length', index_obj.length)
        else:
            # Not all blobs found
            code = http.SERVICE_UNAVAILABLE
            request.setResponseCode(code)
            data = _render_error(code, "Not all blobs present")
            request.setHeader('Content-Type', 'text/html')
            request.setHeader('Content-Length', len(data))
            request.write(data)
    
    def _blob_fetched(self, result, request, session, index_obj):
        blob_id, blob = result
        request.write(blob)
    
    def _blobs_done(self, success, request, session, index_obj):
        log.debug("Finished fetching (success: %i)", success)
        if not success:
            code = http.SERVICE_UNAVAILABLE
            request.setResponseCode(code)
            data = _render_error(code, "Not all blobs present")
            request.setHeader('Content-Type', 'text/html')
            request.setHeader('Content-Length', len(data))
            request.write(data)
        request.finish()
    
    def render_PROPFIND(self, request):
        session, index_obj = self._get_data()
        log.debug("PROPFINDing %s", request.path)
        request_info = self._parse_propfind(request)
        responses = self._find_properties(_get_host(request),
                                          unquote(request.path),
                                          index_obj,
                                          request_info['properties'],
                                          request_info['depth'])
        request.setResponseCode(http.MULTI_STATUS)
        request.setHeader('Content-Type', 'text/xml')
        return _render_template("dav_multistatus.xml",
                                genshi_format='xml',
                                responses=responses,)
    
    def _parse_propfind(self, request):
        # Start result dictionary
        result = {}
        result['depth'] = request.getHeader('Depth')
        if result['depth'] is None:
            result['depth'] = 'infinity'
        # Parse request
        if request.getHeader('Content-Length') == '0':
            result['properties'] = 'all'
        else:
            doc = ElementTree.parse(request.content)
            root = doc.getroot()
            # TODO: Check if root is {DAV:}propfind
            prop_root = root.find('{DAV:}prop')
            # Find properties
            # TODO: Raise error if allprop and propname are both present
            if prop_root is None:
                result['properties'] = 'all'
            elif prop_root.find('{DAV:}allprop') is not None:
                result['properties'] = 'all'
            elif prop_root.find('{DAV:}propname') is not None:
                result['properties'] = 'names'
            else:
                result['properties'] = self._parse_props(prop_root)
        return result
    
    @staticmethod
    def _parse_props(elem):
        result = []
        for prop_elem in elem:
            namespace, name = util.parse_qname(prop_elem.tag)
            new_property = DAVProperty(namespace, name, None)
            result.append(new_property)
        return result
    
    @classmethod
    def _find_properties(cls, host, path, index_obj, properties, depth='0'):
        responses = []
        found, not_found, forbidden = _get_properties(properties, index_obj)
        responses.append(DAVResponse(host + quote(path),
                                     found, not_found, forbidden))
        if index_obj.is_file or depth == '0':
            pass
        else:
            for child in index_obj.get_children():
                child_depth = '0' if depth == '1' else 'infinity'
                result = cls._find_properties(host,
                                              posixpath.join(path, child.name),
                                              child, properties, child_depth)
                responses += result
        return responses
    
    def render_PROPPATCH(self, request):
        session, index_obj = self._get_data()
        request_info = self._parse_proppatch(request)
        locks = _get_locks(request)
        if not index_obj.can_access(locks):
            return _send_error(request, LOCKED, "File is locked")
        properties, forbidden, conflict, failed = \
            _set_properties(request_info, index_obj)
        response = DAVResponse(_get_host(request) + quote(request.path),
                               properties=properties,
                               forbidden=forbidden,
                               conflict=conflict,
                               failed=failed,)
        request.setResponseCode(http.MULTI_STATUS)
        request.setHeader('Content-Type', 'text/xml')
        return _render_template("dav_multistatus.xml",
                                genshi_format='xml',
                                responses=[response],)
    
    def _parse_proppatch(self, request):
        result = []
        # Parse request
        doc = ElementTree.parse(request.content)
        root = doc.getroot()
        # TODO: Check if root is {DAV:}propertyupdate
        for child in root:
            if child.tag == '{DAV:}set':
                prop_elem = child.find('{DAV:}prop')
                result.append(['set'] + self._parse_props(prop_elem))
            elif child.tag == '{DAV:}remove':
                prop_elem = child.find('{DAV:}remove')
                result.append(['remove'] + self._parse_props(prop_elem))
        return result
    
    def render_MKCOL(self, request):
        return _send_error(request, http.NOT_ALLOWED)
    
    def render_DELETE(self, request):
        session, index_obj = self._get_data()
        locks = _get_locks(request)
        if index_obj == index.get_root(session):
            return _send_error(request, http.FORBIDDEN,
                               "Root cannot be deleted")
        else:
            try:
                index_obj.delete(locks, force=True)
            except errors.LockedError:
                return _send_error(request, LOCKED)
            else:
                request.setResponseCode(http.NO_CONTENT)
                return ''
    
    def render_PUT(self, request):
        session, index_obj = self._get_data()
        locks = _get_locks(request)
        if index_obj.is_directory:
            return _send_error(request, http.CONFLICT,
                               "Directories cannot be put")
        path = index_obj.get_path()
        log.debug("Modifying file %s from PUT", path)
        try:
            success = _write_file(session, self.blob_server_list, path,
                                  request.content, locks)
        except errors.LockedError:
            return _send_error(request, LOCKED)
        else:
            if success:
                request.setResponseCode(http.NO_CONTENT)
                return ''
            else:
                return _send_error(request, http.CONFLICT)
    
    def render_COPY(self, request):
        session, index_obj = self._get_data()
        request_info = self._parse_copy(request)
        locks = _get_locks(request)
        host = _get_host(request)
        dest_url = urlparse(request_info['destination'])
        dest_host = dest_url[0] + '://' + dest_url[1]
        dest_path = unquote(dest_url[2])
        # Check for foreign server
        if host != dest_host:
            return _send_error(request, http.BAD_GATEWAY,
                               "This server does not support cross-server "
                               "copying.")
        # Check for root copying
        if index_obj == index.get_root(session):
            return _send_error(request, http.FORBIDDEN,
                               "Root cannot be copied")
        # Check overwriting
        dest_node = index.resolve_path(session, dest_path)
        overwrote = False
        if dest_node is not None:
            if request_info['overwrite']:
                dest_node.delete(locks, force=True)
                overwrote = True
            else:
                return _send_error(request, http.PRECONDITION_FAILED,
                                   "Destination exists")
        # Copy
        try:
            index.move(session, index_obj.get_path(), dest_path, locks,
                       copy=True,
                       recursive=bool(request_info['depth'] == 'infinity'),
                       allow_directory=False)
        except errors.LockedError:
            return _send_error(request, LOCKED)
        except:
            return _send_error(request, http.FORBIDDEN)
        else:
            if overwrote:
                request.setResponseCode(http.NO_CONTENT)
            else:
                request.setResponseCode(http.CREATED)
            return ''
    
    def render_MOVE(self, request):
        session, index_obj = self._get_data()
        request_info = self._parse_copy(request)
        locks = _get_locks(request)
        host = _get_host(request)
        dest_url = urlparse(request_info['destination'])
        dest_host = dest_url[0] + '://' + dest_url[1]
        dest_path = unquote(dest_url[2])
        # Check for foreign server
        if host != dest_host:
            return _send_error(request, http.BAD_GATEWAY,
                               "This server does not support cross-server "
                               "copying.")
        # Check for root copying
        if index_obj == index.get_root(session):
            return _send_error(request, http.FORBIDDEN,
                               "Root cannot be moved")
        # Check for bad depth
        if request_info['depth'] != 'infinity':
            return _send_error(request, http.BAD_REQUEST,
                               "Depth other than infinity given")
        # Check for locks
        if not index_obj.can_access(locks):
            return _send_error(request, LOCKED, "File is locked")
        # Check overwriting
        dest_node = index.resolve_path(session, dest_path)
        overwrote = False
        if dest_node is not None and not dest_node.is_null:
            if request_info['overwrite']:
                dest_node.delete(locks, force=True)
                overwrote = True
            else:
                return _send_error(request, http.PRECONDITION_FAILED,
                                   "Destination exists")
        # Move
        try:
            index.move(session, index_obj.get_path(), dest_path, locks,
                       allow_directory=False)
        except errors.LockedError:
            return _send_error(request, LOCKED)
        except errors.Error:
            return _send_error(request, http.FORBIDDEN)
        else:
            if overwrote:
                request.setResponseCode(http.NO_CONTENT)
            else:
                request.setResponseCode(http.CREATED)
            return ''
    
    def _parse_copy(self, request):
        # Start result dictionary
        result = {}
        result.setdefault('omit', False)
        result.setdefault('keepalive', None)
        # Get headers
        result['depth'] = request.getHeader('Depth')
        if result['depth'] is None:
            result['depth'] = 'infinity'
        overwrite = request.getHeader('Overwrite')
        overwrite_values = {'T': True, 'F': False}
        result['overwrite'] = overwrite_values.get(overwrite, True)
        result['destination'] = request.getHeader('Destination')
        # Parse request (if necessary)
        content_length = request.getHeader('Content-Length')
        if content_length is None or content_length == '0':
            return result
        else:
            doc = ElementTree.parse(request.content)
            root = doc.getroot()
            # Find behavior
            # TODO: Raise error if keepalive and omit are both present
            if root.find('{DAV:}keepalive') is not None:
                keepalive_elem = root.find('{DAV:}keepalive')
                if keepalive_elem.text == '*':
                    result['keepalive'] = '*'
                else:
                    result['keepalive'] = []
                    for child in keepalive_elem:
                        if child.tag == '{DAV:}href':
                            result['keepalive'].append(child.text)
            elif root.find('{DAV:}omit') is not None:
                result['omit'] = True
            return result
    
    def render_LOCK(self, request):
        session, index_obj = self._get_data()
        log.debug("LOCKing %s", request.path)
        request_info = _parse_lock_request(request)
        if isinstance(request_info, (int, long)):
            # Error returned, stop everything
            return _send_error(request, request_info)
        try:
            new_lock = index_obj.lock(exclusive=request_info['exclusive'])
        except errors.LockedError:
            return _send_error(request, LOCKED)
        else:
            log.info("Locked '%s' with token: %s",
                     index_obj.get_path(), new_lock.name)
            request.setResponseCode(http.OK)
            request.setHeader('Content-Type', 'text/xml')
            request.setHeader('Lock-Token', '<%s>' % new_lock.name)
            lock_data = _render_template("lock_discovery.xml",
                                         genshi_format='xml',
                                         comment_trim=True,
                                         locks=index_obj.locks)
            return """\
<?xml version=\"1.0\"?>

<D:prop xmlns:D="DAV:">
    <D:lockdiscovery>
%s
    </D:lockdiscovery>
</D:prop>
""" % lock_data
    
    _lock_token_pattern = re.compile('^<(.*?)>$')
    def render_UNLOCK(self, request):
        session, index_obj = self._get_data()
        # Parse lock token
        lock_token = request.getHeader('Lock-Token')
        match = self._lock_token_pattern.match(lock_token)
        if not match:
            return _send_error(request, http.BAD_REQUEST,
                               "Lock token header malformed")
        lock_token = match.group(1)
        # Unlock
        log.debug("UNLOCKing %s from %r", request.path, lock_token)
        try:
            index_obj.unlock(lock_token)
        except KeyError:
            return _send_error(request, http.PRECONDITION_FAILED,
                               "Lock token not found")
        else:
            log.info("Unlocked '%s' with token: %s",
                     index_obj.get_path(), lock_token)
            if index_obj.is_null and not index_obj.get_lock_status():
                # Delete completely unlocked nulls
                index_obj.delete()
            request.setResponseCode(http.NO_CONTENT)
            return ''

def _parse_chunk(data):
    """
    Parses a single chunk.
    
    :Parameters:
        data : str
            The raw data sent to the channel
    :Returns: ``(chunk, extra_data)``
    :ReturnType: tuple
    """
    CRLF = '\r\n'
    if CRLF in data:
        # We have our first line, parse it
        chunk_line, chunk_data = data.split(CRLF, 1)
        chunk_metadata = chunk_line.split(';')
        chunk_length, chunk_ext = chunk_metadata[0], chunk_metadata[1:]
        chunk_length = int(chunk_length, 16)
        # Now handle data
        if len(chunk_data) >= chunk_length + len(CRLF):
            return (chunk_data[:chunk_length],
                    chunk_data[chunk_length + len(CRLF):])
        else:
            # Not enough data given
            return (None, data)
    else:
        return (None, data)

class HTTPChannel(http.HTTPChannel):
    """Custom HTTP channel that supports chunk encoding."""
    def __init__(self, *args, **kw):
        http.HTTPChannel.__init__(self, *args, **kw)
        self.chunked = 0
        self.chunk_buffer = ''
    
    def lineReceived(self, line):
        self.resetTimeout()
        
        if self.__first_line:
            # if this connection is not persistent, drop any data which
            # the client (illegally) sent after the last request.
            if not self.persistent:
                self.dataReceived = self.lineReceived = lambda *args: None
                return
            
            # IE sends an extraneous empty line (\r\n) after a POST request;
            # eat up such a line, but only ONCE
            if not line and self.__first_line == 1:
                self.__first_line = 2
                return
            
            # create a new Request object
            request = self.requestFactory(self, len(self.requests))
            self.requests.append(request)
            
            self.__first_line = 0
            parts = line.split()
            if len(parts) != 3:
                self.transport.write("HTTP/1.1 400 Bad Request\r\n\r\n")
                self.transport.loseConnection()
                return
            command, request, version = parts
            self._command = command
            self._path = request
            self._version = version
        elif line == '':
            if self.__header:
                self.headerReceived(self.__header)
            self.__header = ''
            self.allHeadersReceived()
            if self.length == 0 and not self.chunked:
                self.allContentReceived()
            else:
                self.setRawMode()
        elif line[0] in ' \t':
            self.__header = self.__header+'\n'+line
        else:
            if self.__header:
                self.headerReceived(self.__header)
            self.__header = line
    
    def headerReceived(self, line):
        # Overridden to nab transfer encoding
        header, data = line.split(':', 1)
        header = header.lower()
        data = data.strip()
        if header == 'transfer-encoding' and data.lower() == 'chunked':
            self.chunked = 1
            # We set the length here at a large number so that request
            # makes a temporary file.  It's an ugly hack, to be sure.
            self.length = 100000
        http.HTTPChannel.headerReceived(self, line)
    
    def allContentReceived(self):
        # Overridden to reset variables
        self.chunked = 0
        self.chunk_buffer = ''
        http.HTTPChannel.allContentReceived(self)
    
    def rawDataReceived(self, data):
        if self.chunked:
            data = self.chunk_buffer + data
            while True:
                chunk, data = _parse_chunk(data)
                if chunk is None:
                    # No more parseable chunks right now
                    break
                elif chunk == '':
                    # Chunk stream finished
                    self.allContentReceived()
                    self.setLineMode(data)
                else:
                    # Everything's normal.  Just write a chunk.
                    self.requests[-1].handleContentChunk(chunk)
            self.chunk_buffer = data
        else:
            http.HTTPChannel.rawDataReceived(self, data)

class ZFNSite(server.Site):
    displayTracebacks = False
    protocol = HTTPChannel
    
    def processingFailed(self, reason):
        log.critical("Unexpected internal server error occurred")
        log.critical("This is probably a bug, and should be reported")
        log.exception(reason.value)
    
    def log(self, request):
        access_log.info('%s - - %s "%s" %d %s "%s" "%s"\n',
                        request.getClientIP(),
                        # request.getUser() or "-",
                        # the remote user is almost never important
                        http._logDateTime,
                        '%s %s %s' % (self._escape(request.method),
                                      self._escape(request.uri),
                                      self._escape(request.clientproto)),
                        request.code,
                        request.sentLength or "-",
                        self._escape(request.getHeader("referer") or "-"),
                        self._escape(request.getHeader("user-agent") or "-"))

def create_server(reactor, engine, blob_server_list,
                  port=DEFAULT_PORT, service_name=None):
    log.debug("Creating WebDAV server...")
    root = ZFNResource(engine, blob_server_list)
    site = ZFNSite(root)
    log.info("Establishing WebDAV Bonjour service...")
    discovery.broadcast(reactor, '_webdav._tcp', port, service_name)
    log.debug("Acquiring port %i...", port)
    reactor.listenTCP(port, site)
