#!/usr/bin/env python
#
#   zfilenet/indexserver/errors.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Index server errors"""

__author__ = 'Ross Light'
__date__ = 'June 28, 2008'
__docformat__ = 'reStructuredText'
__all__ = ['Error',
           'PathExistsError',
           'PathNotFoundError',
           'BadNameError',
           'WrongTypeError',
           'HierarchyError',
           'LockedError',
           'NotSavedError',
           'WrongVersionError',]

class Error(Exception):
    """
    Abstract root class for all index errors.
    
    :CVariables:
        code : int
            The JSON-RPC code to use
    :IVariables:
        path : str
            The path of which the error occurred
        message : str
            The specific message for the error
    """
    code = 0
    
    def __init__(self, path, message=None):
        if path is not None and not isinstance(path, basestring):
            # Probably an index object
            try:
                path = str(path.get_path())
            except AttributeError:
                path = str(path)
        self.path = path
        self.message = message
    
    def _get_message(self):
        """
        Overridden in subclasses to customize the message given.
        
        :Returns: The custom message
        :ReturnType: str
        """
        return None
    
    def __repr__(self):
        if message is None:
            return "%s(%r)" % (type(self).__name__, self.path)
        else:
            return "%s(%r, %r)" % \
                (type(self).__name__, self.path, self.message)
    
    def __str__(self):
        class_message = self._get_message()
        if class_message is not None and self.message is not None:
            return "%s (%s)" % (self._get_message(), self.message)
        elif class_message is not None:
            return class_message
        else:
            return self.message

class PathExistsError(Error):
    """Exception raised when a path already exists and it shouldn't."""
    code = -1
    def _get_message(self):
        return "Path already exists"

class PathNotFoundError(Error):
    """Exception raised when a path doesn't exist and it should."""
    code = -2
    def _get_message(self):
        return "Path does not exist"

class BadNameError(Error):
    """Exception raised when a bad file name is given."""
    code = -3
    def _get_message(self):
        return "Bad file name: %r" % self.path

class WrongTypeError(Error):
    """
    Exception raised when a path is of the wrong type (e.g. a directory's path
    is given when the function expected a file).
    """
    code = -4
    def _get_message(self):
        return "Path is not the correct type"

class HierarchyError(Error):
    """Exception raised when an operation is hierarchially impossible."""
    code = -5
    def _get_message(self):
        return "Hierarchy inconsistent"

class LockedError(Error):
    """Exception raised when the requested file is locked."""
    code = -6
    def _get_message(self):
        return "File is locked"

class NotSavedError(Error):
    """
    Exception raised when an operation requires that an object be saved to the
    database first.
    """
    code = -255
    def _get_message(self):
        return "Object is not in database"

class WrongVersionError(Error):
    """
    Exception raised when the index database is at the wrong version.
    """
    code = -511
    def _get_message(self):
        return "Wrong database version"
