#!/usr/bin/env python
#
#   zfilenet/indexserver/rpcapi.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Index server JSON-RPC API."""

from __future__ import absolute_import, with_statement
from base64 import b64decode, b64encode
import functools
from hashlib import sha1
import logging
from uuid import UUID, uuid4

import pinet
from pinet.dispatch import expose

from . import blobmanager, errors, index
from .. import discovery, util

__author__ = 'Ross Light'
__date__ = 'June 27, 2008'
__all__ = ['DEFAULT_PORT',
           'log',
           'IndexServer',
           'create_server',]

DEFAULT_PORT = 47074
log = logging.getLogger('zfilenet.indexserver.rpcapi')

def handle_index_errors(func):
    """
    Decorator for handling index errors.
    
    :Parameters:
        func
            A function to wrap
    :Returns: The wrapped function
    :ReturnType: function
    """
    def _wrapper_func(*args, **kw):
        try:
            return func(*args, **kw)
        except errors.Error, e:
            error_data = \
            {
                'path': e.path,
                'message': e.message,
            }
            raise pinet.jsonrpc.RPCError(e.code, str(e), error_data)
        except Exception:
            log.exception("Unexpected error occurred during request")
            raise
    # Wrap function attributes
    return functools.update_wrapper(_wrapper_func, func)

class IndexServer(object):
    def __init__(self, reactor, engine, blob_server_list):
        self.reactor = reactor
        self.engine = engine
        self.uploads = {}
        self.blob_server_list = blob_server_list
    
    def _get_session(self):
        return index.create_session(bind=self.engine)
    
    @expose()
    def get_version(self):
        """
        Retrieves the server's version.
        
        :Returns: The server's version string
        :ReturnType: str
        """
        import zfilenet
        return zfilenet.__version__
    
    @expose()
    @handle_index_errors
    def stat(self, path):
        """
        Retrieves the stat table for a path.
        
        :Parameters:
            path : str
                The path to a file
        :Returns: The path's stat table
        :ReturnType: dict
        """
        session = self._get_session()
        node = index.resolve_path(session, path)
        return node.stat() if node is not None else None
    
    @expose()
    @handle_index_errors
    def mkdir(self, path):
        """
        Creates a directory at a given path.
        
        :Parameters:
            path : str
                The path to create
        """
        session = self._get_session()
        # Lock path
        null = index.NullObject.make(session, path)
        with null.lock() as new_lock:
            # Create directory
            index.Directory.make(session, path, [new_lock])
    
    @expose()
    @handle_index_errors
    def listdir(self, path):
        """
        Lists the names in a directory.
        
        :Parameters:
            path : str
                The path to a directory
        :Returns: The names in the directory
        :ReturnType: list
        """
        session = self._get_session()
        node = index.resolve_path(session, path)
        if node is None:
            return None
        elif not node.is_directory:
            raise errors.WrongTypeError(path, "Path is not a directory")
        else:
            return [child.name for child in
                    node.get_children(include_null=False)]
    
    @expose()
    @handle_index_errors
    def delete(self, path):
        """
        Deletes a path.
        
        :Parameters:
            path : str
                The path to delete
        """
        session = self._get_session()
        node = index.resolve_path(session, path)
        if node is None:
            raise errors.PathNotFoundError(path)
        elif node == index.get_root(session):
            raise errors.HierarchyError(path, "Cannot delete root")
        else:
            with node.lock():
                node.delete(force=True)
    
    @expose()
    @handle_index_errors
    def copy(self, src, dst):
        """
        Copies a file or directory.
        
        :Parameters:
            path : str
                The path to copy
        """
        session = self._get_session()
        index.move(session, src, dst, copy=True)
    
    @expose()
    @handle_index_errors
    def move(self, src, dst):
        """
        Moves a file or directory.
        
        :Parameters:
            path : str
                The path to move
        """
        session = self._get_session()
        index.move(session, src, dst)
    
    @expose('download.open', include_client=True)
    @handle_index_errors
    def download_open(self, client, path):
        session = self._get_session()
        node = index.resolve_path(session, path)
        if node is None:
            raise errors.PathNotFoundError(path, "Path does not exist")
        elif not node.is_file:
            raise errors.WrongTypeError(path, "Path is not a file")
        else:
            blob_list = node.blob_list
            file_id = node.uuid
            if not blob_list:
                return dict(file_id=str(file_id),
                            blob_count=0,
                            file_length=0,
                            file_hash=None,)
            else:
                fetcher = self.blob_server_list.fetch_list(blob_list)
                fetcher.progress.addCallback(self._download_blob_fetched,
                                             file_id, client)
                fetcher.finished.addCallback(self._download_done,
                                             file_id, client)
                # We will start this on the next event loop.
                # Why?  Because if we have no blob servers, we want the failure
                # to be spit at the client after getting info.
                self.reactor.callLater(0.0, fetcher.start)
                return dict(file_id=str(file_id),
                            blob_count=len(blob_list),
                            file_length=node.length,
                            file_hash=util.data2hex(node.sha1_hash),)
    
    def _download_blob_fetched(self, result, file_id, client):
        blob_id, blob = result
        client.notify('download.receive_data', file_id.hex, b64encode(blob))
    
    def _download_done(self, success, file_id, client):
        client.notify('download.close', file_id.hex, success)
    
    @expose('upload.open')
    @handle_index_errors
    def upload_open(self, path):
        session = self._get_session()
        node = index.resolve_path(session, path)
        if node is None:
            node = index.File.make(session, path)
        elif not node.is_file:
            raise errors.WrongTypeError(path, "Path is not a file")
        else:
            if node.can_access([]):
                node.clear()
            else:
                raise errors.LockedError(path, "Path is locked")
        new_lock = node.lock()
        writer = self.blob_server_list.get_writer()
        self.uploads[node.uuid] = dict(blobs=[], sha1_hash=sha1(), length=0,
                                       writer=writer,)
        return dict(upload_id=str(node.uuid),
                    lock=new_lock.name,)
    
    @expose('upload.write')
    @handle_index_errors
    def upload_write(self, file_id, data, lock=None):
        session = self._get_session()
        file_id = util.to_uuid(file_id)
        try:
            upload = self.uploads[file_id]
        except KeyError:
            raise pinet.jsonrpc.RPCError(1, "Upload does not exist")
        # Check for lock
        locks = [lock] if lock is not None else []
        upload_file = index.IndexObject.by_uuid(session, file_id)
        if not upload_file.can_access(locks, none_ignore=False):
            raise errors.LockedError(upload_file, "Upload lock not present")
        # Write blob to blob servers
        new_blob_id = uuid4()
        decoded_data = b64decode(data)
        upload['sha1_hash'].update(decoded_data)
        upload['length'] += len(decoded_data)
        upload['blobs'].append(new_blob_id)
        # TODO: Split into chunks
        upload['writer'].write_blob(new_blob_id, decoded_data)
    
    @expose('upload.close')
    @handle_index_errors
    def upload_close(self, file_id, lock=None):
        session = self._get_session()
        file_id = util.to_uuid(file_id)
        try:
            upload = self.uploads.pop(file_id)
        except KeyError:
            raise pinet.jsonrpc.RPCError(1, "Upload does not exist")
        upload_file = index.IndexObject.by_uuid(session, file_id)
        # Check for lock
        locks = [lock] if lock is not None else []
        if not upload_file.can_access(locks, none_ignore=False):
            raise errors.LockedError(upload_file, "Upload lock not present")
        # Finish upload
        upload_file.length = upload['length']
        upload_file.sha1_hash = upload['sha1_hash'].digest()
        upload_file.blob_list = upload['blobs']
        if lock is not None:
            upload_file.unlock(lock)
        session.commit()

def create_server(reactor, engine, blob_server_list,
                  port=DEFAULT_PORT, service_name=None):
    log.debug("Creating JSON-RPC API server...")
    peer = IndexServer(reactor, engine, blob_server_list)
    factory = pinet.jsonrpc.JSON_RPCFactory()
    factory.requests.register_object(peer)
    log.info("Establishing JSON-RPC Bonjour service...")
    discovery.broadcast(reactor, discovery.INDEX_SERVICE_TYPE,
                        port, service_name)
    log.debug("Acquiring port %i...", port)
    reactor.listenTCP(port, factory)
